"""Displays currently being selected device(s).
"""

import sys
import json
import time
from PyQt5 import QtWidgets, QtCore

from beamline_console.logger.logit import Logit
from beamline_console.devices.motordevice import MotorBasedDevice
from beamline_console.snapshot.slot import Slot
from beamline_console.devicecontrol.motoritemwidget import MotorItemWidget

from beamline_console.gui.DeviceWidget_ui import Ui_DeviceWidget

# ----------------------------------------------------------------------
class DeviceWidget(QtWidgets.QWidget, Logit):
    """
    """
    save_snapshot = QtCore.pyqtSignal(str)  # motors selection info

    DATETIME = "%Y/%m/%d %H:%M:%S"

    # ----------------------------------------------------------------------
    def __init__(self, parent):
        """
        Args:
            parent (QWidget)
        """

        QtWidgets.QWidget.__init__(self, parent)
        Logit.__init__(self)

        self._ui = Ui_DeviceWidget()
        self._ui.setupUi(self)

        self._selected_devices = []
        self.widgets_map = {}

        self._beamline_hal = None
        self._settings = None

        self._device_widget_layout = QtWidgets.QVBoxLayout(self._ui.wi_device_container)

        self._timer = QtCore.QTimer(self)
        self._timer.timeout.connect(self.refresh_gui)

        self._ui.but_save.clicked.connect(self._save_snapshot)
        self._ui.but_stop.clicked.connect(self.stop_motors)

    # ----------------------------------------------------------------------
    def initialize(self, beamline_hal, settings):
        """
        """
        self._beamline_hal = beamline_hal
        self._settings = settings

        for dev_name, dev_handle in sorted(self._beamline_hal.device_map.items()):
            for motor_name, motor_handle in dev_handle.motors.items():
                self.widgets_map[f"{dev_name}/{motor_name}"] = MotorItemWidget(self._beamline_hal, dev_handle, motor_handle, self)

        self._timer.start(int(self._settings.option("beamline_ui", "selected_device_refresh_period")))

        self.logger.info("DeviceWidget initialized successfully")

    # ----------------------------------------------------------------------
    def refresh_gui(self):
        """Synchronize GUI (beamline state can be changed by Jive, Astor, etc.)
        """
        for device_name in self._selected_devices:
            try:
                device = self._beamline_hal.device(device_name)
                if isinstance(device, MotorBasedDevice):
                    for motor in device.list_motors():
                        self.widgets_map[f"{device.name}/{motor}"].refresh_gui()
                elif isinstance(device, Slot):
                    for motor in device.list_motors():
                        self.widgets_map[motor].refresh_gui()
            except KeyError:
                pass
            except Exception as e:
                self.logger.error(f"DeviceWidget: cannot refresh gui: {e}")

    # ----------------------------------------------------------------------
    def stop_motors(self):
        """
        """
        try:
            self._beamline_hal.group_action("stop_motors")
            time.sleep(0.5)

            self.sync_gui("absolute")

        except Exception as err:
            self.logger.error(err, exc_info=sys.exc_info())
            raise

            # non motor devs TODO

        self.logger.info("All motors stopped")

    # ----------------------------------------------------------------------
    def selected_device_changed(self, selected_devices):
        """Called when a device is selected in the BeamlineWidget.
        """

        layout = self._device_widget_layout

        for i in reversed(list(range(layout.count()))):
            item = layout.itemAt(i)
            if item:
                widget = layout.itemAt(i).widget()
                if widget:
                    layout.removeWidget(widget)
                    widget.setVisible(False)
                else:
                    layout.removeItem(item)

        self._selected_devices = selected_devices

        for name in selected_devices:
            device = self._beamline_hal.device(name)
            for motor_name in device.list_motors():
                if isinstance(device, MotorBasedDevice):
                    motor_name = f"{device.name}/{motor_name}"
                self.widgets_map[motor_name].setVisible(True)
                layout.addWidget(self.widgets_map[motor_name])
            spacer = QtWidgets.QSpacerItem(22, 20, QtWidgets.QSizePolicy.Expanding,
                                           QtWidgets.QSizePolicy.Expanding)
            layout.addSpacerItem(spacer)

        self.sync_gui("absolute")

    # ----------------------------------------------------------------------
    def _save_snapshot(self):
        """
        """
        # support for non-motor devices? TODO
        self.save_snapshot.emit(json.dumps({"devices": self._selected_devices}))

    # ----------------------------------------------------------------------
    def sync_gui(self, mode="absolute"):
        """
        """
        for device in self.widgets_map.values():
            for motor in device:
                motor.sync_pos(mode, "async")



"""Manage DeviceGraphicsItem objects representing beamline devices.
This class is tightly coupled with BeamlineGraphicsView.
"""

import os

from PyQt5.QtWidgets import QGraphicsScene, QGraphicsPixmapItem
from PyQt5.QtCore import pyqtSignal, QRectF, Qt
from PyQt5.QtGui import QPixmap
from distutils.util import strtobool
from beamline_console.logger.logit import Logit
from beamline_console.devicecontrol.deviceitem.devicegraphicsitem import DeviceGraphicsItem


# ----------------------------------------------------------------------
class BeamlineGraphicsScene(QGraphicsScene, Logit):
    """
    """
    device_selection_changed = pyqtSignal(list)

    # ----------------------------------------------------------------------
    def __init__(self, parent, beamline_hal, settings):

        self._device_list = list(beamline_hal.device_map.keys())

        widget_rows = int(settings.option("beamline_ui", "widget_rows"))
        scene_height = int(settings.option("beamline_ui", "section_height"))

        slot_mode = strtobool(settings.option("general", "enable_slots"))

        if slot_mode:
            main_folder = os.path.join(os.path.expanduser('~'), '.beamline_console')
            background_file = os.path.join(main_folder, settings.option("background", "background_file"))
            if not os.path.isfile(background_file):
                raise RuntimeError(f"Cannot open background file {background_file}")
            pixmap = QPixmap(background_file)
            scene_width = scene_height * pixmap.width()/pixmap.height()
        else:
            scene_width = len(self._device_list) // widget_rows
            scene_width += 1 if len(self._device_list) % widget_rows else 0
            scene_width *= scene_height

        QGraphicsScene.__init__(self, QRectF(0, 0, scene_width, scene_height), parent)
        Logit.__init__(self)

        self._selected_devices = []
        self.device_items = {}

        if slot_mode:
            pixmap_item = QGraphicsPixmapItem(pixmap)
            self.addItem(pixmap_item)

            for slot_name, slot_item in beamline_hal.slots_map.items():
                self.device_items[slot_name] = DeviceGraphicsItem(self, slot_item.name, slot_item.position, True)

            for device_name, device_item in beamline_hal.device_map.items():
                if device_item.widget_position is not None:
                    self.device_items[device_name] = DeviceGraphicsItem(self, device_item.name, device_item.widget_position)

            for item in list(self.device_items.values()):
                item.set_active(False)

        else:
            horizontal_step = int(settings.option("beamline_ui", "horizontal_step"))
            vertical_step = int(settings.option("beamline_ui", "vertical_step"))
            device_width = int(settings.option("beamline_ui", "device_width"))
            device_height = int(settings.option("beamline_ui", "device_height"))

            shift_x = (horizontal_step - device_width) // 2
            shift_y = (vertical_step - device_height) // 2

            row_count = 0
            column_count = 0
            for device in self._device_list:
                geometry = [shift_x + column_count * horizontal_step,
                            shift_y + row_count * vertical_step, device_width, device_height]
                self.device_items[device] = DeviceGraphicsItem(self, device, geometry, False)
                row_count += 1
                if row_count == widget_rows:
                    row_count = 0
                    column_count += 1
   
    # ----------------------------------------------------------------------
    def mouseMoveEvent(self, event):
        """
        """
        super().mouseMoveEvent(event)
        
        #pos = event.scenePos()

    # ----------------------------------------------------------------------
    def mousePressEvent(self, event):
        """
        """
        super().mousePressEvent(event)

        pos = event.scenePos()

        if Qt.ControlModifier != event.modifiers():
            self._selected_devices = []
            
        device_name = self._intersect(pos)
        if device_name:
            if device_name not in self._selected_devices:
                self._selected_devices.append(device_name)
            else:
                self._selected_devices.remove(device_name)

        self.logger.debug(f"BeamlineGraphicsScene: new selection: {self._selected_devices}")
        self.device_selection_changed.emit(self._selected_devices)

    # ----------------------------------------------------------------------
    def _intersect(self, pos):
        """
        """
        for name, graphics_item in list(self.device_items.items()):
            if graphics_item.intersect(pos):
                return name

        return None



# yury.matveev@desy.de

from beamline_console.logger.logit import Logit


class Slot(Logit):

    # ----------------------------------------------------------------------
    def __init__(self, device_name, beamline_hal):
        """
        """
        super().__init__()
        self.name = device_name
        self.beamline_hal = beamline_hal
        self.logger.debug(f"Instantiating device object {device_name}")
        self.position = []
        self.snapshots = {}
        self.motors_list = []
        self.xml_node = None

        # ----------------------------------------------------------------------
    def from_xml(self, xml_node):
        """
            Load device's configuration from a given XML node.
        """
        general_access = xml_node.getAttribute('access')
        if not general_access:
            general_access = "user"

        widget = xml_node.getElementsByTagName("widget")[0]
        self.position = [int(widget.getAttribute("x")), int(widget.getAttribute("y")),
                         int(widget.getAttribute("width")), int(widget.getAttribute("height"))]

        self.xml_node = xml_node

    # ----------------------------------------------------------------------
    def list_motors(self):
        return self.motors_list

    # ----------------------------------------------------------------------
    def initialize(self):
        """
        """
        for snapshot_node in self.xml_node.getElementsByTagName("snapshot"):
            snapshot_name = str(snapshot_node.getAttribute(f"name"))
            motor_found = True
            counter = 1
            while motor_found:
                motor_name = str(snapshot_node.getAttribute(f"motor{counter}"))
                if motor_name == "":
                    motor_found = False
                    continue
                if motor_name not in self.beamline_hal.list_motors():
                    self.logger.error(f"Slot {self.name}: snapshot {snapshot_name}: motor {motor_name} not defined in BeamlineHal")
                    return False

                if motor_name not in self.motors_list:
                    self.motors_list.append(motor_name)

                counter += 1

        return True



"""Logger propagating log records to the LoggingWidget.
"""

from logging import StreamHandler
from PyQt5 import QtCore


# ----------------------------------------------------------------------
class GuiLogger(StreamHandler):

    def __init__(self):
        StreamHandler.__init__(self)
        self.logger = Logger()

    # ----------------------------------------------------------------------
    def emit(self, record):
        try:
            self.logger.emit_record.emit(int(record.levelno), f'{record.asctime.split(",")[0]} {record.levelname} {record.msg}')
        except:
            print(f'{int(record.levelno)}: {record.asctime.split(",")[0]} {record.levelname} {record.msg}')


# ----------------------------------------------------------------------
class Logger(QtCore.QObject):

    emit_record = QtCore.pyqtSignal(int, str)
